@@ -12,16 +12,24 @@
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
 # obtain one at https://mozilla.org/MPL/2.0/.
 
+"""Note this reference page is very much WIP.
+See `Hypothesis for the scientific stack
+<https://hypothesis.readthedocs.io/en/latest/numpy.html#numpy>`_
+for a general idea on what we're doing here.
+"""
+
 import math
+from functools import update_wrapper
 from types import SimpleNamespace
-from typing import (Any, Iterable, List, Mapping, Optional, Sequence, Tuple,
-                    Type, TypeVar, Union)
+from typing import (Any, Iterable, List, Mapping, NamedTuple, Optional,
+                    Sequence, Tuple, Type, TypeVar, Union)
 from warnings import warn
 
 from hypothesis import strategies as st
 from hypothesis.errors import HypothesisWarning, InvalidArgument
 from hypothesis.internal.conjecture import utils as cu
 from hypothesis.internal.validation import check_type, check_valid_interval
+from hypothesis.strategies._internal.strategies import check_strategy
 
 __all__ = [
     "get_strategies_namespace",
@@ -34,6 +42,8 @@ __all__ = [
     "unsigned_integer_dtypes",
     "floating_dtypes",
     "valid_tuple_axes",
+    "broadcastable_shapes",
+    "mutually_broadcastable_shapes",
 ]
 
 
@@ -45,6 +55,12 @@ DataType = Union[Boolean, SignedInteger, UnsignedInteger, Float]
 Array = TypeVar("Array")
 Shape = Tuple[int, ...]
 
+
+class BroadcastableShapes(NamedTuple):
+    input_shapes: Tuple[Shape, ...]
+    result_shape: Shape
+
+
 DTYPE_NAMES = [
     "bool",
     "int8", "int16", "int32", "int64",
@@ -108,26 +124,10 @@ def order_check(name, floor, min_, max_):
     if min_ > max_:
         raise InvalidArgument(f"min_{name}={min_} is larger than max_{name}={max_}")
 
-
-def get_strategies_namespace(xp) -> SimpleNamespace:
-    infer_xp_is_compliant(xp)
-
-    return SimpleNamespace(
-        from_dtype=lambda *a, **kw: from_dtype(xp, *a, **kw),
-        arrays=lambda *a, **kw: arrays(xp, *a, **kw),
-        array_shapes=array_shapes,
-        scalar_dtypes=lambda *a, **kw: scalar_dtypes(xp, *a, **kw),
-        boolean_dtypes=lambda *a, **kw: boolean_dtypes(xp, *a, **kw),
-        integer_dtypes=lambda *a, **kw: integer_dtypes(xp, *a, **kw),
-        unsigned_integer_dtypes=lambda *a, **kw: unsigned_integer_dtypes(xp, *a, **kw),
-        floating_dtypes=lambda *a, **kw: floating_dtypes(xp, *a, **kw),
-        valid_tuple_axes=valid_tuple_axes,
-    )
-
-
 # Note NumPy supports non-array scalars which hypothesis.extra.numpy.from_dtype
 # utilises, but this from_dtype() method returns just base strategies.
 
+
 def from_dtype(
     xp,
     dtype: Union[Type[DataType], str],
@@ -139,6 +139,8 @@ def from_dtype(
     exclude_min: Optional[bool] = None,
     exclude_max: Optional[bool] = None,
 ) -> st.SearchStrategy[Union[bool, int, float]]:
+    """Creates a strategy which can generate any castable value of the given
+    Array API dtype."""
     infer_xp_is_compliant(xp)
 
     if isinstance(dtype, str):
@@ -346,6 +348,8 @@ def arrays(
     fill: Optional[st.SearchStrategy[Any]] = None,
     unique: bool = False,
 ) -> st.SearchStrategy[Array]:
+    """Returns a strategy for generating Array API arrays."""
+
     infer_xp_is_compliant(xp)
     check_xp_attributes(xp, ["empty", "full", "all", "isnan", "bool", "reshape"])
 
@@ -361,16 +365,23 @@ def arrays(
     if isinstance(shape, int):
         shape = (shape,)
 
+    if not all(isinstance(s, int) for s in shape):
+        raise InvalidArgument(
+            f"Array shape must be integer in each dimension, provided shape was {shape}"
+        )
+
     if elements is None:
         elements = from_dtype(xp, dtype)
     elif isinstance(elements, Mapping):
         elements = from_dtype(xp, dtype, **elements)
+    check_strategy(elements, "elements")
 
     if fill is None:
         if unique or not elements.has_reusable_values:
             fill = st.nothing()
         else:
             fill = elements
+    check_strategy(fill, "fill")
 
     return ArrayStrategy(xp, elements, dtype, shape, fill, unique)
 
@@ -382,6 +393,7 @@ def array_shapes(
     min_side: int = 1,
     max_side: Optional[int] = None,
 ) -> st.SearchStrategy[Shape]:
+    """Return a strategy for array shapes (tuples of int >= 1)."""
     check_type(int, min_dims, "min_dims")
     check_type(int, min_side, "min_side")
 
@@ -401,11 +413,6 @@ def array_shapes(
     ).map(tuple)
 
 
-# We assume there are dtypes objects part of the array module namespace.
-# Note there is a current discussion about whether this is expected behaviour:
-# github.com/data-apis/array-api/issues/152
-
-
 def check_dtypes(xp, dtypes: List[Type[DataType]], stubs: List[str]):
     if len(dtypes) == 0:
         f_stubs = ", ".join(stubs)
@@ -418,6 +425,7 @@ def check_dtypes(xp, dtypes: List[Type[DataType]], stubs: List[str]):
 
 
 def scalar_dtypes(xp) -> st.SearchStrategy[Type[DataType]]:
+    """Return a strategy for all Array API dtypes."""
     infer_xp_is_compliant(xp)
 
     dtypes, stubs = partition_attributes_and_stubs(xp, DTYPE_NAMES)
@@ -459,8 +467,9 @@ def numeric_dtype_names(base_name: str, sizes: Sequence[int]):
 
 
 def integer_dtypes(
-    xp, sizes: Union[int, Sequence[int]] = (8, 16, 32, 64)
+    xp, *, sizes: Union[int, Sequence[int]] = (8, 16, 32, 64)
 ) -> st.SearchStrategy[Type[SignedInteger]]:
+    """Return a strategy for signed integer dtypes in the Array API."""
     infer_xp_is_compliant(xp)
 
     if isinstance(sizes, int):
@@ -476,8 +485,9 @@ def integer_dtypes(
 
 
 def unsigned_integer_dtypes(
-    xp, sizes: Union[int, Sequence[int]] = (8, 16, 32, 64)
+    xp, *, sizes: Union[int, Sequence[int]] = (8, 16, 32, 64)
 ) -> st.SearchStrategy[Type[UnsignedInteger]]:
+    """Return a strategy for unsigned integer dtypes in the Array API."""
     infer_xp_is_compliant(xp)
 
     if isinstance(sizes, int):
@@ -493,8 +503,9 @@ def unsigned_integer_dtypes(
 
 
 def floating_dtypes(
-    xp, sizes: Union[int, Sequence[int]] = (32, 64)
+    xp, *, sizes: Union[int, Sequence[int]] = (32, 64)
 ) -> st.SearchStrategy[Type[Float]]:
+    """Return a strategy for floating dtypes in the Array API."""
     infer_xp_is_compliant(xp)
 
     if isinstance(sizes, int):
@@ -515,6 +526,7 @@ def valid_tuple_axes(
     min_size: int = 0,
     max_size: Optional[int] = None,
 ) -> st.SearchStrategy[Shape]:
+    """Return a strategy for something I don't quite understand yet."""
     if max_size is None:
         max_size = ndim
 
@@ -531,3 +543,270 @@ def valid_tuple_axes(
     return st.lists(
         axes, min_size=min_size, max_size=max_size, unique_by=lambda x: x % ndim
     ).map(tuple)
+
+
+class MutuallyBroadcastableShapesStrategy(st.SearchStrategy):
+    def __init__(
+        self,
+        num_shapes,
+        base_shape=(),
+        min_dims=0,
+        max_dims=None,
+        min_side=1,
+        max_side=None,
+    ):
+        assert 0 <= min_side <= max_side
+        assert 0 <= min_dims <= max_dims
+
+        self.base_shape = base_shape
+        self.num_shapes = num_shapes
+        self.min_dims = min_dims
+        self.max_dims = max_dims
+        self.min_side = min_side
+        self.max_side = max_side
+
+        self.side_strat = st.integers(min_side, max_side)
+        self.size_one_allowed = self.min_side <= 1 <= self.max_side
+
+    def do_draw(self, data):
+        # All shapes are handled in column-major order; i.e. they are reversed
+        base_shape = self.base_shape[::-1]
+        result_shape = list(base_shape)
+        shapes = [[] for _ in range(self.num_shapes)]
+        use = [True for _ in range(self.num_shapes)]
+
+        for dim_count in range(1, self.max_dims + 1):
+            dim = dim_count - 1
+
+            # We begin by drawing a valid dimension-size for the given
+            # dimension. This restricts the variability across the shapes
+            # at this dimension such that they can only choose between
+            # this size and a singleton dimension.
+            if len(base_shape) < dim_count or base_shape[dim] == 1:
+                # dim is unrestricted by the base-shape: shrink to min_side
+                dim_side = data.draw(self.side_strat)
+            elif base_shape[dim] <= self.max_side:
+                # dim is aligned with non-singleton base-dim
+                dim_side = base_shape[dim]
+            else:
+                # only a singleton is valid in alignment with the base-dim
+                dim_side = 1
+
+            for shape_id, shape in enumerate(shapes):
+                # Populating this dimension-size for each shape, either
+                # the drawn size is used or, if permitted, a singleton
+                # dimension.
+                if dim_count <= len(base_shape) and self.size_one_allowed:
+                    # aligned: shrink towards size 1
+                    side = data.draw(st.sampled_from([1, dim_side]))
+                else:
+                    side = dim_side
+
+                # Use a trick where where a biased coin is queried to see
+                # if the given shape-tuple will continue to be grown. All
+                # of the relevant draws will still be made for the given
+                # shape-tuple even if it is no longer being added to.
+                # This helps to ensure more stable shrinking behavior.
+                if self.min_dims < dim_count:
+                    use[shape_id] &= cu.biased_coin(
+                        data, 1 - 1 / (1 + self.max_dims - dim)
+                    )
+
+                if use[shape_id]:
+                    shape.append(side)
+                    if len(result_shape) < len(shape):
+                        result_shape.append(shape[-1])
+                    elif shape[-1] != 1 and result_shape[dim] == 1:
+                        result_shape[dim] = shape[-1]
+            if not any(use):
+                break
+
+        result_shape = result_shape[: max(map(len, [self.base_shape] + shapes))]
+
+        assert len(shapes) == self.num_shapes
+        assert all(self.min_dims <= len(s) <= self.max_dims for s in shapes)
+        assert all(self.min_side <= s <= self.max_side for side in shapes for s in side)
+
+        return BroadcastableShapes(
+            input_shapes=tuple(tuple(reversed(shape)) for shape in shapes),
+            result_shape=tuple(reversed(result_shape)),
+        )
+
+
+def broadcastable_shapes(
+    shape: Shape,
+    *,
+    min_dims: int = 0,
+    max_dims: Optional[int] = None,
+    min_side: int = 1,
+    max_side: Optional[int] = None,
+) -> st.SearchStrategy[Shape]:
+    """Return a strategy for generating shapes that are broadcast-compatible
+    with the provided shape."""
+    check_type(tuple, shape, "shape")
+    strict_check = max_side is None or max_dims is None
+    check_type(int, min_side, "min_side")
+    check_type(int, min_dims, "min_dims")
+
+    if max_dims is None:
+        max_dims = min(32, max(len(shape), min_dims) + 2)
+    else:
+        check_type(int, max_dims, "max_dims")
+
+    if max_side is None:
+        max_side = max(tuple(shape[-max_dims:]) + (min_side,)) + 2
+    else:
+        check_type(int, max_side, "max_side")
+
+    order_check("dims", 0, min_dims, max_dims)
+    order_check("side", 0, min_side, max_side)
+
+    if 32 < max_dims:
+        raise InvalidArgument("max_dims cannot exceed 32")
+
+    dims, bound_name = (max_dims, "max_dims") if strict_check else (
+        min_dims, "min_dims")
+
+    # check for unsatisfiable min_side
+    if not all(min_side <= s for s in shape[::-1][:dims] if s != 1):
+        raise InvalidArgument(
+            "Given shape=%r, there are no broadcast-compatible "
+            "shapes that satisfy: %s=%s and min_side=%s"
+            % (shape, bound_name, dims, min_side)
+        )
+
+    # check for unsatisfiable [min_side, max_side]
+    if not (
+        min_side <= 1 <= max_side or all(s <= max_side for s in shape[::-1][:dims])
+    ):
+        raise InvalidArgument(
+            "Given shape=%r, there are no broadcast-compatible shapes "
+            "that satisfy: %s=%s and [min_side=%s, max_side=%s]"
+            % (shape, bound_name, dims, min_side, max_side)
+        )
+
+    if not strict_check:
+        # reduce max_dims to exclude unsatisfiable dimensions
+        for n, s in zip(range(max_dims), reversed(shape)):
+            if s < min_side and s != 1:
+                max_dims = n
+                break
+            elif not (min_side <= 1 <= max_side or s <= max_side):
+                max_dims = n
+                break
+
+    return MutuallyBroadcastableShapesStrategy(
+        num_shapes=1,
+        base_shape=shape,
+        min_dims=min_dims,
+        max_dims=max_dims,
+        min_side=min_side,
+        max_side=max_side,
+    ).map(lambda x: x.input_shapes[0])
+
+
+def mutually_broadcastable_shapes(
+    num_shapes: int,
+    *,
+    base_shape: Shape = (),
+    min_dims: int = 0,
+    max_dims: Optional[int] = None,
+    min_side: int = 1,
+    max_side: Optional[int] = None,
+) -> st.SearchStrategy[BroadcastableShapes]:
+    """Return a strategy for generating a specified number of shapes
+    ``num_shapes`` that are mutually-broadcastable with one another and with the
+    provided base shape ``base_shape``."""
+
+    check_type(int, num_shapes, "num_shapes")
+    if num_shapes < 1:
+        raise InvalidArgument(f"num_shapes={num_shapes} must be at least 1")
+
+    check_type(tuple, base_shape, "base_shape")
+    check_type(int, min_side, "min_side")
+    check_type(int, min_dims, "min_dims")
+
+    strict_check = max_dims is not None
+
+    if max_dims is None:
+        max_dims = min(32, max(len(base_shape), min_dims) + 2)
+    check_type(int, max_dims, "max_dims")
+
+    if max_side is None:
+        max_side = max(base_shape[-max_dims:] + (min_side,)) + 2
+    check_type(int, max_side, "max_side")
+
+    order_check("dims", 0, min_dims, max_dims)
+    order_check("side", 0, min_side, max_side)
+
+    if strict_check:
+        dims = max_dims
+        bound_name = "max_dims"
+    else:
+        dims = min_dims
+        bound_name = "min_dims"
+
+    # check for unsatisfiable min_side
+    if not all(min_side <= s for s in base_shape[::-1][:dims] if s != 1):
+        raise InvalidArgument(
+            f"Given base_shape={base_shape}, there are no broadcast-compatible"
+            f" shapes that satisfy: {bound_name}={dims} and min_side={min_side}"
+        )
+
+    # check for unsatisfiable [min_side, max_side]
+    if not (
+        min_side <= 1 <= max_side or all(s <= max_side for s in base_shape[::-1][:dims])
+    ):
+        raise InvalidArgument(
+            f"Given base_shape={base_shape}, there are no broadcast-compatible"
+            f" shapes that satisfy all of {bound_name}={dims},"
+            f" min_side={min_side}, and max_side={max_side}"
+        )
+
+    if not strict_check:
+        # reduce max_dims to exclude unsatisfiable dimensions
+        for n, s in zip(range(max_dims), base_shape[::-1]):
+            if s < min_side and s != 1:
+                max_dims = n
+                break
+            elif not (min_side <= 1 <= max_side or s <= max_side):
+                max_dims = n
+                break
+
+    return MutuallyBroadcastableShapesStrategy(
+        num_shapes=num_shapes,
+        base_shape=base_shape,
+        min_dims=min_dims,
+        max_dims=max_dims,
+        min_side=min_side,
+        max_side=max_side,
+    )
+
+
+def get_strategies_namespace(xp) -> SimpleNamespace:
+    """Creates a strategies namespace."""
+    infer_xp_is_compliant(xp)
+
+    attributes = {
+        "from_dtype": lambda *a, **kw: from_dtype(xp, *a, **kw),
+        "arrays": lambda *a, **kw: arrays(xp, *a, **kw),
+        "array_shapes": array_shapes,
+        "scalar_dtypes": lambda *a, **kw: scalar_dtypes(xp, *a, **kw),
+        "boolean_dtypes": lambda *a, **kw: boolean_dtypes(xp, *a, **kw),
+        "integer_dtypes": lambda *a, **kw: integer_dtypes(xp, *a, **kw),
+        "unsigned_integer_dtypes": lambda *a, **kw: unsigned_integer_dtypes(
+            xp, *a, **kw
+        ),
+        "floating_dtypes": lambda *a, **kw: floating_dtypes(xp, *a, **kw),
+        "valid_tuple_axes": valid_tuple_axes,
+        "broadcastable_shapes": broadcastable_shapes,
+        "mutually_broadcastable_shapes": mutually_broadcastable_shapes,
+    }
+
+    globals_ = globals()
+    for name in attributes.keys():
+        orig_func = globals_[name]
+        wrapped_func = attributes[name]
+        attributes[name] = update_wrapper(wrapped_func, orig_func)
+
+    return SimpleNamespace(**attributes)
